from anticrlf.exception import UnsafeSubstitutionError


class SubstitutionMap(dict):
    """
    SubstitutionMap is a special-purpose ``dict`` for CRLF substitution maps

    ``SubstitutionMap`` is meant to provide error resistance when customizing replacements in
    the ``anticrlf`` log formatter.

    By default, an "empty" ``SubstitutionMap`` will map "\n" to "\\n" (linefeed to the string
    '\n') and "\r" to "\\r" (carriage return to the string '\r')

    Attempting to assign a value that contains any of the keys raises
    ``UnsafeSubstitutionError``, since setting a key is essentially claiming that key is an
    unsafe string

    Deleting either the "\n" or "\r" keys (e.g. using ``del``) resets them to their default
    values
    """
    def __init__(self, **kwargs):
        super(self.__class__, self).__init__(**kwargs)

        # if the constructor didn't provide substitutions for CR and LF, set up the defaults
        if "\n" not in self.keys():
            super(self.__class__, self).__setitem__("\n", "\\n")
        if "\r" not in self.keys():
            super(self.__class__, self).__setitem__("\r", "\\r")

        self.check_value()

    def __setitem__(self, key, value):
        self.check_value(key, value)
        super(self.__class__, self).__setitem__(key, value)

    def __delitem__(self, key):
        super(self.__class__, self).__delitem__(key)
        if "\n" not in self.keys():
            super(self.__class__, self).__setitem__("\n", "\\n")
        if "\r" not in self.keys():
            super(self.__class__, self).__setitem__("\r", "\\r")

    def check_value(self, key=None, value=None):
        subvalues = list(self.values())
        subkeys = list(self.keys())
        if value is not None and value not in subvalues:
            subvalues.append(value)
        if key is not None and key not in subkeys:
            subkeys.append(key)

        for val in subvalues:
            # if any value contains any key, throw an error; since the keys are unsafe strings, this prevents
            # a developer from accidentally setting a value that contains a string considered unsafe, at least
            # without trapping the errors -- can't stop determined sabotage...
            for key in subkeys:
                if key in val:
                    raise UnsafeSubstitutionError("Cannot assign a substitution that contains a value declared unsafe")



""" a ``logging`` Formatter that escapes newline chars to avoid CRLF log injection (CWE-93)

Defines the class ``AntiCrlfFormatter``

"""
from __future__ import unicode_literals
import logging
import warnings

from anticrlf.types import SubstitutionMap


class LogFormatter(logging.Formatter):
    """logging Formatter to sanitize CRLF errors (CWE-93)

    This class is a drop-in replacement for ``logging.Formatter``, and has the
    exact same construction arguments. However, as a final step of formatting a
    log line, it escapes carriage returns (\r) and linefeeds (\n).

    By default, these are replaced with their escaped equivalents (see `Examples`_),
    but the ``replacements`` dictionary can be modified to change this behabior.

    Examples:
        ::

            import anticrlf

            handler = logging.StreamHandler(sys.stderr)
            handler.setFormatter(anticrlf.LogFormatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s'))

            logger = logging.getLogger(__name__)
            logger.addHandler(handler)
            logger.setLevel(logging.INFO)

            logger.info("Example text with a newline\nhere")

        This results in::

            2017-02-03 08:43:52,557 - __main__ - INFO - Example text with a newline\nhere

        Whereas with the default ``Formatter``, it would be::

            2017-02-03 08:43:52,557 - __main__ - INFO - Example text with a newline
            here

        If you wanted newlines to be replaced with \x0A instead, you could::

            formatter = anticrlf.LogFormatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
            formatter.replacements["\n"] = "\\x0A"  # Note the double backslash for literal!
            handler.setFormatter(formatter)

    """
    def __init__(self, fmt=None, datefmt=None):
        super(self.__class__, self).__init__(fmt=fmt, datefmt=datefmt)
        self.replacements = SubstitutionMap()  # defaults to mapping \n: \\n and \r: \\r

    def format(self, record):
        """calls logger.Formatter.format, then removes CR and LF from the resulting message before returning it"""
        if type(self.replacements) != SubstitutionMap:
            warnings.warn(UserWarning("replacements invalid: resetting to defaults"))
            self.replacements = SubstitutionMap()

        formatted_message = super(self.__class__, self).format(record)

        for repl in self.replacements:
            formatted_message = formatted_message.replace(repl, self.replacements[repl])

        return formatted_message



from setuptools import setup

from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.rst')) as f:
    long_description = f.read()

setup(
    name='logging-formatter-anticrlf',
    version='1.2',
    packages=['anticrlf'],
    url='https://github.com/darrenpmeyer/logging-formatter-anticrlf',
    license='BSD 2-clause',
    author='Darren P Meyer',
    author_email='darren@darrenpmeyer.com',
    description='Python logging Formatter for CRLF Injection (CWE-93 / CWE-117) prevention',
    long_description=long_description,
    long_description_content_type='text/x-rst'
)
